﻿/*
http://www.sno.phy.queensu.ca/~phil/exiftool/
*/

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Text;
using System.Threading;
using System.Linq;

namespace BBCSharp
{
    public sealed class ExifToolWrapper : IDisposable
    {
        public string Exe { get; private set; }
        private const string ExeName = "exiftool(-k).exe";
        private const string Arguments = "-fast -m -q -q -stay_open True -@ - -common_args -d \"%Y.%m.%d %H:%M:%S\" -c \"%d %d %.6f\" -t";   //-g for groups

        public enum Statuses { Stopped, Starting, Ready, Stopping };
        public Statuses Status { get; private set; }

        public bool Resurrect { get; set; }
        private bool stopRequested = false;

        private int cmdCnt = 1;
        private readonly StringBuilder output = new StringBuilder();
        private readonly ProcessStartInfo psi;
        private Process proc = null;
        private readonly ManualResetEvent waitHandle = new ManualResetEvent(true);

        public string ExiftoolVersion { get; private set; }

        public ExifToolWrapper(string path = null)
        {
            Resurrect = true;

            Exe = string.IsNullOrEmpty(path) ? Path.Combine(Path.GetDirectoryName(Assembly.GetEntryAssembly().Location), ExeName) : path;
            if(!File.Exists(Exe))
                throw new ExifToolException(ExeName + " not found");

            psi = new ProcessStartInfo
            {
                FileName = Exe,
                Arguments = Arguments,
                CreateNoWindow = true,
                UseShellExecute = false,
                RedirectStandardOutput = true,
                RedirectStandardError = true,
                RedirectStandardInput = true
            };

            Status = Statuses.Stopped;
        }

        private void OutputDataReceived(object sender, DataReceivedEventArgs e)
        {
            if(string.IsNullOrEmpty(e.Data))
                return;

            if(Status == Statuses.Starting)
            {
                ExiftoolVersion = e.Data;
                waitHandle.Set();
            }
            else
            {
                if(e.Data.ToLower() == string.Format("{{ready{0}}}", cmdCnt))
                    waitHandle.Set();
                else
                    output.AppendLine(e.Data);
            }
        }

        public void Start()
        {
            stopRequested = false;

            if(Status != Statuses.Stopped)
                throw new ExifToolException("Process is not stopped");

            Status = Statuses.Starting;

            proc = new Process { StartInfo = psi, EnableRaisingEvents = true };
            proc.OutputDataReceived += OutputDataReceived;
            proc.Exited += proc_Exited;
            proc.Start();

            proc.BeginOutputReadLine();

            waitHandle.Reset();
            proc.StandardInput.WriteLine("-ver\n-execute0000");
            waitHandle.WaitOne();

            Status = Statuses.Ready;
        }

        //detect if process is killed
        void proc_Exited(object sender, EventArgs e)
        {
            if(proc != null)
            {
                proc.Dispose();
                proc = null;
            }

            Status = Statuses.Stopped;

            waitHandle.Set();

            if(!stopRequested && Resurrect)
                Start();
        }

        public void Stop()
        {
            stopRequested = true;

            if(Status != Statuses.Ready)
                throw new ExifToolException("Process must be ready");

            Status = Statuses.Stopping;

            waitHandle.Reset();
            proc.StandardInput.WriteLine("-stay_open\nFalse\n");
            if(!waitHandle.WaitOne(5000))
            {
                if(proc != null)
                {
                    //silently swallow an eventual exception
                    try
                    {
                        proc.Kill();
                        proc.WaitForExit(2000);
                        proc.Dispose();
                    }
                    catch { }

                    proc = null;
                }

                Status = Statuses.Stopped;
            }
        }

        private static readonly Object lockObj = new Object();

        public string SendCommand(string cmd, params object[] args)
        {
            if(Status != Statuses.Ready)
                throw new ExifToolException("Process must be ready");

            string r;
            lock(lockObj)
            {
                waitHandle.Reset();
                proc.StandardInput.WriteLine("{0}\n-execute{1}", args.Length == 0 ? cmd : string.Format(cmd, args), cmdCnt);
                waitHandle.WaitOne();

                cmdCnt++;

                r = output.ToString();
                output.Clear();
            }

            return r;
        }

        public bool SetExifInto(string path, Dictionary<string, string> data, bool overwriteOriginal = true)
        {
            if(!File.Exists(path))
                return false;

            StringBuilder cmd = new StringBuilder();
            foreach(KeyValuePair<string, string> kv in data)
            {
                cmd.AppendFormat("-{0}={1}\n", kv.Key, kv.Value);
            }
            
            if(overwriteOriginal)
                cmd.AppendLine("-overwrite_original");
            
            cmd.Append(path);
            string sRes = SendCommand(cmd.ToString());

            return sRes.Contains("1 image files updated");
        }

        public Dictionary<string, string> FetchExifFrom(string path, IEnumerable<string> tagsToKeep = null, bool keepKeysWithEmptyValues = true)
        {
            Dictionary<string, string> res = new Dictionary<string, string>();

            bool filter = tagsToKeep != null && tagsToKeep.Any();
            string sRes = SendCommand(path);
            foreach(string s in sRes.Split(new [] { Environment.NewLine }, StringSplitOptions.RemoveEmptyEntries))
            {
                string[] kv = s.Split('\t');
                Debug.Assert(kv.Length == 2, "Can not parse line :'" + s + "'");

                if(kv.Length != 2 || (!keepKeysWithEmptyValues && string.IsNullOrEmpty(kv[1])))
                    continue;

                if(filter && !tagsToKeep.Contains(kv[0]))
                    continue;
                    
                res[kv[0]] = kv[1];
            }

            return res;
        }

        public List<string> FetchExifToListFrom(string path, IEnumerable<string> tagsToKeep = null, bool keepKeysWithEmptyValues = true, string separator = ": ")
        {
            List<string> res = new List<string>();

            bool filter = tagsToKeep != null && tagsToKeep.Any();
            string sRes = SendCommand(path);
            foreach(string s in sRes.Split(new[] { Environment.NewLine }, StringSplitOptions.RemoveEmptyEntries))
            {
                string[] kv = s.Split('\t');
                Debug.Assert(kv.Length == 2, "Can not parse line :'" + s + "'");

                if(kv.Length != 2 || (!keepKeysWithEmptyValues && string.IsNullOrEmpty(kv[1])))
                    continue;

                if(filter && !tagsToKeep.Contains(kv[0]))
                    continue;

                res.Add(string.Format("{0}{1}{2}", kv[0], separator, kv[1]));
            }

            return res;
        }

        public bool CloneExif(string source, string dest, bool backup = false)
        {
            if(!File.Exists(source) || !File.Exists(dest))
                return false;

            string sRes = SendCommand("{0}-tagsFromFile\n{1}\n{2}", backup ? "" : "-overwrite_original\n", source, dest);

            return sRes.Contains("1 image files updated");
        }

        public DateTime? GetCreationTime(string path)
        {
            DateTime dt;
            if(DateTime.TryParseExact(SendCommand("-DateTimeOriginal\n-s3\n{0}", path), 
                "yyyy.MM.dd HH:mm:ss", 
                CultureInfo.InvariantCulture,
                DateTimeStyles.AllowWhiteSpaces, 
                out dt))
                return dt;

            return null;
        }

        public int GetOrientation(string path)
        {
            int o;
            if(Int32.TryParse(SendCommand("-Orientation\n-n\n-s3\n{0}", path).Trim(new[] { '\t', '\r', '\n' }), out o))
                return o;

            return 1;
        }

        public int GetOrientationDeg(string path)
        {
            return OrientationPos2Deg(GetOrientation(path));
        }

        public bool SetOrientation(string path, int ori, bool overwriteOriginal = true)
        {
            if(!File.Exists(path))
                return false;

            StringBuilder cmd = new StringBuilder();
            cmd.AppendFormat("-Orientation={0}\n-n\n-s3\n", ori);

            if(overwriteOriginal)
                cmd.AppendLine("-overwrite_original");

            cmd.Append(path);
            string sRes = SendCommand(cmd.ToString());

            return sRes.Contains("1 image files updated");
        }

        public bool SetOrientationDeg(string path, int ori, bool overwriteOriginal = true)
        {
            return SetOrientation(path, OrientationDeg2Pos(ori), overwriteOriginal);
        }

        #region Static helpers

        /*
         
1        2       3      4         5            6           7          8

888888  888888      88  88      8888888888  88                  88  8888888888
88          88      88  88      88  88      88  88          88  88      88  88
8888      8888    8888  8888    88          8888888888  8888888888          88
88          88      88  88
88          88  888888  888888
                                                      at least 6 and 8 seems to be inverted
         */

        //TODO: maybe some kind of map would be better
        public static int OrientationPos2Deg(int pos)
        {
            switch(pos)
            {
                case 8:
                    return 270;
                case 3:
                    return 180;
                case 6:
                    return 90;
                default:
                    return 0;
            }
        }

        public static int OrientationDeg2Pos(int deg)
        {
            switch(deg)
            {
                case 270:
                    return 8;
                case 180:
                    return 3;
                case 90:
                    return 6;
                default:
                    return 1;
            }
        }

        /*
        1 => 'Horizontal (normal)',
        2 => 'Mirror horizontal',
        3 => 'Rotate 180',
        4 => 'Mirror vertical',
        5 => 'Mirror horizontal and rotate 270 CW',
        6 => 'Rotate 90 CW',
        7 => 'Mirror horizontal and rotate 90 CW',
        8 => 'Rotate 270 CW',
        */

        public static int OrientationString2Deg(string pos)
        {
            switch(pos)
            {
                case "Rotate 270 CW":
                    return 270;
                case "Rotate 180":
                    return 180;
                case "Rotate 90 CW":
                    return 90;
                default:
                    return 0;
            }
        }

        public static string OrientationDeg2String(int deg)
        {
            switch(deg)
            {
                case 270:
                    return "Rotate 270 CW";
                case 180:
                    return "Rotate 180";
                case 90:
                    return "Rotate 90 CW";
                default:
                    return "Horizontal (normal)";
            }
        }

        private static readonly int[] orientationPositions = { 1, 6, 3, 8 };

        public static int RotateOrientation(int crtOri, bool clockwise, int steps = 1)
        {
            int newOri = 1;
            int len = orientationPositions.Length;

            if(steps % len == 0)
                return crtOri;

            for(int i = 0; i < len; i++)
            {
                if(crtOri == orientationPositions[i])
                {
                    newOri = clockwise
                        ? orientationPositions[(i + steps) % len]
                        : orientationPositions[(i + (1 + steps / len) * len - steps) % orientationPositions.Length];

                    break;
                }
            }

            return newOri;
        }

        #endregion
        
        #region IDisposable Members

        public void Dispose()
        {
            Debug.Assert(Status == Statuses.Ready || Status == Statuses.Stopped, "Invalid state");

            if(proc != null && Status == Statuses.Ready)
                Stop();

            waitHandle.Dispose();
        }

        #endregion
    }
}
